
use core::fmt::Write;

use tinyvec::*;

#[cfg(feature = "bytemuck")]
use bytemuck::*;

use chlorine::pick;
pick! {
  if #[cfg(windows)] {
    mod windows;
    use windows as sys;
    type SysTerm = sys::WindowsTerminal;
  } else if #[cfg(unix)] {
    mod unix;
    use unix as sys;
    type SysTerm = sys::UnixTerminal;
  } else {
    compile_error!("This crate doesn't support this system.")
  }
}

pub const ESC: u8 = 0x1B;
pub const CTRL_Q: u8 = ctrl_key(b'q');
pub const CTRL_C: u8 = ctrl_key(b'c');

/// Converts a letter byte into the byte for that Control key.
///
/// eg: ctrl_key(b'c') -> 3_u8
pub const fn ctrl_key(b: u8) -> u8 {
  b & 0b0001_1111
}

#[derive(Debug, Clone, Copy)]
#[repr(C)]
pub struct WinSize {
  pub rows: u16,
  pub cols: u16,
}
#[cfg(feature = "bytemuck")]
unsafe impl Zeroable for WinSize {}
#[cfg(feature = "bytemuck")]
unsafe impl Pod for WinSize {}

#[derive(Debug, Clone, Copy)]
pub enum Key {
  Lit(u8),
  F(u8),
  Escape,
  Tab,
  Enter,
  ArrowLeft,
  ArrowUp,
  ArrowRight,
  ArrowDown,
  Home,
  End,
  Insert,
  Delete,
  Backspace,
  PageUp,
  PageDown,
  Keypad5,
}

#[derive(Clone, Copy, PartialEq, Eq)]
pub struct Mods(pub u8);
#[allow(bad_style)]
impl Mods {
  pub const None: Mods = Mods(0);
  pub const Meta: Mods = Mods(0b1000);
  pub const Ctrl: Mods = Mods(0b0100);
  pub const Alt: Mods = Mods(0b0010);
  pub const Shift: Mods = Mods(0b0001);
}
impl core::fmt::Debug for Mods {
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    if self.0 == Self::None.0 {
      write!(f, "None")
    } else {
      if self.0 & Self::Meta.0 > 0 {
        write!(f, "Meta,")?;
      }
      if self.0 & Self::Ctrl.0 > 0 {
        write!(f, "Ctrl,")?;
      }
      if self.0 & Self::Alt.0 > 0 {
        write!(f, "Alt,")?;
      }
      if self.0 & Self::Shift.0 > 0 {
        write!(f, "Shift,")?;
      }
      Ok(())
    }
  }
}

#[derive(Debug, Clone, Copy)]
pub enum TermEvent {
  Resize { rows: u16, cols: u16 },
  Key { val: Key, mods: Mods },
  Weirdness { buf: ArrayVec<[u8; 8]> },
}

#[repr(transparent)]
pub struct Terminal(SysTerm);

impl Terminal {
  #[allow(clippy::new_without_default)]
  pub fn new() -> Self {
    Self(SysTerm::new())
  }

  pub fn flush(&mut self) -> std::io::Result<()> {
    self.0.flush()
  }

  pub fn poll_events(&mut self) -> Result<Option<TermEvent>, i32> {
    self.0.poll_events().map_err(|e|e.raw_os_error().unwrap_or(7))
  }
}

impl core::fmt::Write for Terminal {
  fn write_str(&mut self, s: &str) -> Result<(), core::fmt::Error> {
    self.0.write_str(s)
  }
}
