#![cfg(unix)]

use super::*;

use core::{
  mem::ManuallyDrop,
  sync::atomic::{AtomicBool, Ordering},
};
use core::mem::{transmute, zeroed};
use std::io::{
  BufWriter, Read, Stderr, StderrLock, Stdin, StdinLock, Stdout, StdoutLock, stdin, stderr, stdout,
};

pub use libc::c_int;

/// Used by the `win_size_change_handler` function.
///
/// When this is flipped to `true`, the window size has changed and you should
/// re-check the window size.
pub static TERM_SIZE_OLD: AtomicBool = AtomicBool::new(false);

/// Flips `TERM_SIZE_OLD` to `true` with `Relaxed` ordering.
///
/// Ignores all arguments, they are simply to match the required signature.
pub extern "C" fn win_size_change_handler(
  _sig: libc::c_int,
  _info: *const libc::siginfo_t,
  _context: *const core::ffi::c_void,
) {
  TERM_SIZE_OLD.store(true, Ordering::Relaxed);
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum RegisterError {
  CouldNotRegister,
  AlreadyRegistered,
}

/// Registers the window size changed handler.
/// ## Safety
/// You agree to only call this once per process.
fn register_win_size_change_handler() -> Result<(), RegisterError> {
  /// If we've registered the handler before or not.
  static HANDLER_REGISTERED: AtomicBool = AtomicBool::new(false);

  if !HANDLER_REGISTERED.compare_and_swap(false, true, Ordering::SeqCst) {
    let mut new_act: libc::sigaction = unsafe { zeroed() };
    new_act.sa_sigaction = win_size_change_handler as usize;
    new_act.sa_flags = libc::SA_SIGINFO;
    if 0
      == unsafe {
        libc::sigaction(libc::SIGWINCH, &new_act, core::ptr::null_mut())
      }
    {
      Ok(())
    } else {
      Err(RegisterError::CouldNotRegister)
    }
  } else {
    Err(RegisterError::AlreadyRegistered)
  }
}

fn read_error_number() -> c_int {
  // Safety: the error number (`errno`) is just a thread local `c_int`.
  // However, in C it's generally provided via CPP macro, and then the exact
  // details can vary wildly by platform.
  pick! {
    if #[cfg(target_os = "vxworks")] {
      // get the value directly, no pointer!
      unsafe { libc::errnoGet() }
    } else if #[cfg(any(target_os = "linux", target_os = "emscripten", target_os = "redox"))] {
      // here and below get a raw pointer which you then read from.
      unsafe { *libc::__errno_location() }
    } else if #[cfg(any(target_os = "netbsd", target_os = "openbsd", target_os = "android"))] {
      unsafe { *libc::__errno() }
    } else if #[cfg(any(target_os = "solaris", target_os = "illumos"))] {
      unsafe { *libc::___errno() }
    } else if #[cfg(any(target_os = "macos", target_os = "freebsd"))] {
      unsafe { *libc::__error() }
    } else if #[cfg(target_os = "haiku")] {
      unsafe { *libc::_errnop() }
    } else if #[cfg(target_os = "dragonfly")] {
      // blocked on https://github.com/rust-lang/rust/issues/29594
      -1
    } else {
      // you're using an unknown unix flavor, sorry!
      -1
    }
  }
}

/// Gets the window size of the terminal connected to `stdout`
///
/// ## Failure
/// * If this fails you get the C `errno` value instead.
/// * The call is most likely to fail because `stdout` just isn't a tty device,
///   such as because the program output is being piped elsewhere.
#[inline]
pub fn ioctl_get_window_size() -> Result<WinSize, c_int> {
  use libc::{ioctl, winsize, STDOUT_FILENO, TIOCGWINSZ};

  let mut ws: winsize = unsafe { core::mem::zeroed() };
  let ioctl_ret = unsafe { ioctl(STDOUT_FILENO, TIOCGWINSZ, &mut ws) };
  if ioctl_ret == 0 {
    Ok(WinSize { rows: ws.ws_row, cols: ws.ws_col })
  } else {
    Err(read_error_number())
  }
}

/// Gets the window size of the terminal that stdout is connected to.
pub fn get_window_size() -> Option<WinSize> {
  // This is the preferred execution path, it's much faster.
  if let Ok(size) = ioctl_get_window_size() {
    return Some(size);
  }

  use core::str::from_utf8;
  use std::process::Command;

  let output = Command::new("tput").arg("lines").output().ok()?;
  let rows: u16 = if output.status.success() {
    from_utf8(&output.stdout).ok()?.parse().ok()?
  } else {
    return None;
  };

  let output = Command::new("tput").arg("cols").output().ok()?;
  let cols: u16 = if output.status.success() {
    from_utf8(&output.stdout).ok()?.parse().ok()?
  } else {
    return None;
  };

  Some(WinSize { rows, cols })
}

pub struct UnixTerminal {
  orig_term: libc::termios,

  std_in: ManuallyDrop<Stdin>,
  std_in_lock: ManuallyDrop<StdinLock<'static>>,
  input_progress: ArrayVec<[u8; 8]>,

  std_out: ManuallyDrop<Stdout>,
  std_out_buf: ManuallyDrop<BufWriter<StdoutLock<'static>>>,

  std_err: ManuallyDrop<Stderr>,
  std_err_buf: ManuallyDrop<BufWriter<StderrLock<'static>>>,
}
impl Drop for UnixTerminal {
  fn drop(&mut self) {
    unsafe {
      use libc::{tcsetattr, STDIN_FILENO, TCSAFLUSH};
      tcsetattr(STDIN_FILENO, TCSAFLUSH, &self.orig_term);
      ManuallyDrop::drop(&mut self.std_in_lock);
      ManuallyDrop::drop(&mut self.std_in);
      ManuallyDrop::drop(&mut self.std_out_buf);
      ManuallyDrop::drop(&mut self.std_out);
      ManuallyDrop::drop(&mut self.std_err_buf);
      ManuallyDrop::drop(&mut self.std_err);
    }
  }
}
impl core::fmt::Debug for UnixTerminal {
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    write!(f, "UnixTerminal {{ ... }}")
  }
}
impl UnixTerminal {
  pub fn new() -> Self {
    let std_in = stdin();
    let std_out = stdout();
    let std_err = stderr();
    let std_in_lock: StdinLock<'static> = unsafe { transmute(std_in.lock()) };
    let std_out_lock: StdoutLock<'static> =
      unsafe { transmute(std_out.lock()) };
    let std_err_lock: StderrLock<'static> =
      unsafe { transmute(std_err.lock()) };
    let std_out_buf = BufWriter::new(std_out_lock);
    let std_err_buf = BufWriter::new(std_err_lock);

    register_win_size_change_handler().expect("failed to register the signal handler");

    use libc::{
      tcgetattr, tcsetattr, BRKINT, CS8, ECHO, ICANON, ICRNL, IEXTEN, INPCK,
      ISIG, ISTRIP, IXON, OPOST, STDIN_FILENO, TCSAFLUSH, VMIN, VTIME,
    };
    let mut orig_term: libc::termios = unsafe { zeroed() };
    unsafe {
      if tcgetattr(STDIN_FILENO, &mut orig_term) != 0 {
        panic!("Couldn't read attributes");
      }
    }
    let mut raw = orig_term;
    raw.c_iflag &= !(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
    raw.c_oflag &= !OPOST;
    raw.c_cflag |= CS8;
    raw.c_lflag &= !(ECHO | ICANON | ISIG | IEXTEN);
    raw.c_cc[VMIN] = 0;
    raw.c_cc[VTIME] = 1;
    unsafe {
      if tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw) != 0 {
        panic!("Couldn't write attributes.")
      }
    }
    Self {
      orig_term,
      std_in: ManuallyDrop::new(std_in),
      std_in_lock: ManuallyDrop::new(std_in_lock),
      input_progress: ArrayVec::default(),
      std_out: ManuallyDrop::new(std_out),
      std_out_buf: ManuallyDrop::new(std_out_buf),
      std_err: ManuallyDrop::new(std_err),
      std_err_buf: ManuallyDrop::new(std_err_buf),
    }
  }

  pub fn flush(&mut self) -> std::io::Result<()> {
    use std::io::Write;
    self.std_out_buf.flush()
  }

  /// Read into the buffer.
  ///
  /// Returns Ok(bool) for if the buffer got some data or not,
  /// or Err(e) for any IO error _other than_ an Interrupted error.
  /// In the case of an Interrupted error, it will eat the error and
  /// try again.
  fn poll_for_bytes(&mut self) -> Result<usize, std::io::Error> {
    use std::io::ErrorKind;
    loop {
      match self.std_in_lock.read(self.input_progress.grab_spare_slice_mut()) {
        Ok(count) => {
          self.input_progress.set_len(self.input_progress.len() + count);
          return Ok(count);
        }
        Err(e) => {
          if e.kind() == ErrorKind::Interrupted {
            // Note(Lokathor): The window-size-changed handler can fire while
            // we're reading, in which case we hit this branch. Since we might
            // have some state partly read and also some in the buffer we need
            // to finish off this event and return it, then the next call to
            // poll_events can respond to the signal handler and give the
            // window size changed event to the caller. For now we continue
            // as if the window size hasn't actually changed yet.
            continue;
          } else {
            return Err(e);
          }
        }
      }
    }
  }

  /// Polls for any pending events.
  ///
  /// * Resize events are given top priority.
  /// * Otherwise the terminal will grab input and do its best to interpret
  ///   what's found.
  pub fn poll_events(&mut self) -> Result<Option<TermEvent>, std::io::Error> {
    // At the start of event polling, give priority to a window resize.
    if TERM_SIZE_OLD.compare_and_swap(true, false, Ordering::Relaxed) {
      let WinSize { rows, cols } = get_window_size().unwrap();
      return Ok(Some(TermEvent::Resize { rows, cols }));
    }

    // Possible input sequences should be one of:
    // * <non-escape-byte>                     .standard key press
    // * <esc> None                            .Escape key
    // * <esc> <esc>                           .also the Escape key
    // * <esc> <char>                          ."Alt keypress or keycode sequence"
    // * <esc> [                               .Alt+[ (other keys seem to also use this for alt+key?)
    // * <esc> O LETTER_KEY_CODE               .Some of the function keys, some of the time

    // * <esc> [ LETTER_CODE                   .unmodified keycode (eg: `ESC[H`)
    // * <esc> [ NUM_? ; NUM_MODS LETTER_CODE  .modified keycode (eg: `ESC[1;5D`)

    // * <esc> [ NUM_CODE ~                    .unmodified numcode (eg: `ESC[1~`)
    // * <esc> [ NUM_CODE ; NUM_MODS ~         .modified numcode (eg: `ESC[3;5~`)

    if self.input_progress.len() < 8 {
      self.poll_for_bytes()?;
    }

    // Ok(Some(TermEvent::Key { val: Key::Escape, mods: Mods::None }))

    if self.input_progress.is_empty() {
      return Ok(None);
    }

    if cfg!(debug_assertions) {
      // Note(Lokathor): We do this weird direct usage of the buffer because of borrowing
      // nonsense.
      use std::io::Write;
      write!(self.std_out_buf, "{{. ").unwrap();
      for b in self.input_progress.iter() {
        match *b {
          ESC => write!(self.std_out_buf, "ESC").unwrap(),
          9 => write!(self.std_out_buf, "TAB").unwrap(),
          13 => write!(self.std_out_buf, "ENT").unwrap(),
          127 => write!(self.std_out_buf, "BKSP").unwrap(),
          other if other.is_ascii_control() => {
            write!(self.std_out_buf, "0x{:02X}", other).unwrap()
          }
          other => write!(self.std_out_buf, "{}", other as char).unwrap(),
        }
        write!(self.std_out_buf, " ").unwrap();
      }
      write!(self.std_out_buf, ".}}\r\n").unwrap();
    }

    match self.input_progress.remove(0) {
      ESC if self.input_progress.is_empty() => {
        // ESC and a timeout for more means that it was probably Escape key
        return Ok(Some(TermEvent::Key { val: Key::Escape, mods: Mods::None }));
      }
      ESC => (),
      9 => return Ok(Some(TermEvent::Key { val: Key::Tab, mods: Mods::None })),
      13 => {
        return Ok(Some(TermEvent::Key { val: Key::Enter, mods: Mods::None }))
      }
      127 => {
        return Ok(Some(TermEvent::Key {
          val: Key::Backspace,
          mods: Mods::None,
        }))
      }
      other => {
        return Ok(Some(TermEvent::Key {
          val: Key::Lit(other),
          mods: Mods::None,
        }))
      }
    }

    let after_b_esc_or_empty =
      self.input_progress.len() == 1 || self.input_progress[1] == ESC;
    match self.input_progress.remove(0) {
      ESC => {
        // two ESC in a row counts as a single press of the escape key.
        return Ok(Some(TermEvent::Key { val: Key::Escape, mods: Mods::None }));
      }
      9 => return Ok(Some(TermEvent::Key { val: Key::Tab, mods: Mods::Alt })),
      13 => {
        return Ok(Some(TermEvent::Key { val: Key::Enter, mods: Mods::Alt }))
      }
      127 => {
        return Ok(Some(TermEvent::Key {
          val: Key::Backspace,
          mods: Mods::Alt,
        }))
      }
      b'O' if !after_b_esc_or_empty => match self.input_progress.remove(0) {
        b'P' => {
          return Ok(Some(TermEvent::Key { val: Key::F(1), mods: Mods::None }))
        }
        b'Q' => {
          return Ok(Some(TermEvent::Key { val: Key::F(2), mods: Mods::None }))
        }
        b'R' => {
          return Ok(Some(TermEvent::Key { val: Key::F(3), mods: Mods::None }))
        }
        b'S' => {
          return Ok(Some(TermEvent::Key { val: Key::F(4), mods: Mods::None }))
        }
        b'T' => {
          return Ok(Some(TermEvent::Key { val: Key::F(5), mods: Mods::None }))
        }
        b'U' => {
          return Ok(Some(TermEvent::Key { val: Key::F(6), mods: Mods::None }))
        }
        b'V' => {
          return Ok(Some(TermEvent::Key { val: Key::F(7), mods: Mods::None }))
        }
        b'W' => {
          return Ok(Some(TermEvent::Key { val: Key::F(8), mods: Mods::None }))
        }
        b'X' => {
          return Ok(Some(TermEvent::Key { val: Key::F(9), mods: Mods::None }))
        }
        b'Y' => {
          return Ok(Some(TermEvent::Key { val: Key::F(10), mods: Mods::None }))
        }
        b'Z' => {
          return Ok(Some(TermEvent::Key { val: Key::F(11), mods: Mods::None }))
        }
        b'[' => {
          return Ok(Some(TermEvent::Key { val: Key::F(12), mods: Mods::None }))
        }
        c => {
          let buf = array_vec!([u8; 8], ESC, b'0', c);
          return Ok(Some(TermEvent::Weirdness { buf }));
        }
      },
      b'[' if !after_b_esc_or_empty => {
        let mut num_code = 0;
        let mut key_mods = 0;
        let mut c = self.input_progress.remove(0);
        while (c as char).is_digit(10) {
          num_code *= 10;
          num_code += c - b'0';
          c = self.input_progress.remove(0);
        }
        if c == b';' {
          c = self.input_progress.remove(0);
          while (c as char).is_digit(10) {
            key_mods *= 10;
            key_mods += c - b'0';
            c = self.input_progress.remove(0);
          }
        }
        let mods = if key_mods > 0 { Mods(key_mods - 1) } else { Mods::None };
        match c {
          b'~' => match num_code {
            1 => return Ok(Some(TermEvent::Key { val: Key::Home, mods })),
            2 => return Ok(Some(TermEvent::Key { val: Key::Insert, mods })),
            3 => return Ok(Some(TermEvent::Key { val: Key::Delete, mods })),
            4 => return Ok(Some(TermEvent::Key { val: Key::End, mods })),
            5 => return Ok(Some(TermEvent::Key { val: Key::PageUp, mods })),
            6 => return Ok(Some(TermEvent::Key { val: Key::PageDown, mods })),
            15 => return Ok(Some(TermEvent::Key { val: Key::F(5), mods })),
            17 => return Ok(Some(TermEvent::Key { val: Key::F(6), mods })),
            18 => return Ok(Some(TermEvent::Key { val: Key::F(7), mods })),
            19 => return Ok(Some(TermEvent::Key { val: Key::F(8), mods })),
            20 => return Ok(Some(TermEvent::Key { val: Key::F(9), mods })),
            21 => return Ok(Some(TermEvent::Key { val: Key::F(10), mods })),
            23 => return Ok(Some(TermEvent::Key { val: Key::F(11), mods })),
            24 => return Ok(Some(TermEvent::Key { val: Key::F(12), mods })),
            25 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(1),
                mods: Mods::Shift,
              }))
            }
            26 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(2),
                mods: Mods::Shift,
              }))
            }
            28 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(3),
                mods: Mods::Shift,
              }))
            }
            29 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(4),
                mods: Mods::Shift,
              }))
            }
            31 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(5),
                mods: Mods::Shift,
              }))
            }
            32 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(6),
                mods: Mods::Shift,
              }))
            }
            33 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(7),
                mods: Mods::Shift,
              }))
            }
            34 => {
              return Ok(Some(TermEvent::Key {
                val: Key::F(8),
                mods: Mods::Shift,
              }))
            }
            _ => todo!(
              "unknown num {} mods {} spare {:?}",
              num_code,
              key_mods,
              self.input_progress
            ),
          },
          b'A' => return Ok(Some(TermEvent::Key { val: Key::ArrowUp, mods })),
          b'B' => {
            return Ok(Some(TermEvent::Key { val: Key::ArrowDown, mods }))
          }
          b'C' => {
            return Ok(Some(TermEvent::Key { val: Key::ArrowRight, mods }))
          }
          b'D' => {
            return Ok(Some(TermEvent::Key { val: Key::ArrowLeft, mods }))
          }
          b'E' => return Ok(Some(TermEvent::Key { val: Key::Keypad5, mods })),
          b'F' => return Ok(Some(TermEvent::Key { val: Key::End, mods })),
          b'G' => return Ok(Some(TermEvent::Key { val: Key::Keypad5, mods })),
          b'H' => return Ok(Some(TermEvent::Key { val: Key::Home, mods })),
          b'P' => return Ok(Some(TermEvent::Key { val: Key::F(1), mods })),
          b'Q' => return Ok(Some(TermEvent::Key { val: Key::F(2), mods })),
          b'R' => return Ok(Some(TermEvent::Key { val: Key::F(3), mods })),
          b'S' => return Ok(Some(TermEvent::Key { val: Key::F(4), mods })),
          b'[' => {
            let f = self.input_progress.remove(0);
            return Ok(Some(TermEvent::Key { val: Key::F(f - 64), mods }));
          }
          _ => todo!(
            "unknown terminus b{:?}, num {} mods {} spare {:?}",
            (c as char),
            num_code,
            key_mods,
            self.input_progress
          ),
        }
      }
      other => {
        // `ESC key` is Alt+key, including Alt+[ or Alt+o if nothing followed.
        return Ok(Some(TermEvent::Key {
          val: Key::Lit(other),
          mods: Mods::Alt,
        }));
      }
    }
  }
}

impl Write for UnixTerminal {
  #[inline]
  fn write_str(&mut self, s: &str) -> Result<(), core::fmt::Error> {
    use std::io::Write;
    self.std_out_buf.write(s.as_bytes())
      .map(|_|())
      .map_err(|_|core::fmt::Error)
  }
}
