#![cfg(windows)]

use super::*;

use core::fmt::Write;
use core::mem::{transmute, ManuallyDrop};
use std::io::{
  stderr, stdin, stdout, Stderr, StderrLock, Stdin, StdinLock,
  Stdout, StdoutLock,
};

use fenestroj::{
  consoleapi::{
    get_console_mode, read_console_input, set_console_mode, write_console,
  },
  processenv::{get_std_handle, StdHandle},
  wincon::{
    DISABLE_NEWLINE_AUTO_RETURN, ENABLE_VIRTUAL_TERMINAL_PROCESSING, ENABLE_PROCESSED_OUTPUT,
    ENABLE_WINDOW_INPUT,
  },
  wincontypes::{
    INPUT_RECORD, KEY_EVENT, LEFT_ALT_PRESSED, LEFT_CTRL_PRESSED,
    RIGHT_ALT_PRESSED, RIGHT_CTRL_PRESSED, SHIFT_PRESSED,
    WINDOW_BUFFER_SIZE_EVENT,
  },
  winnt::HANDLE,
};

pub struct WindowsTerminal {
  orig_console_input: u32,
  orig_console_output: u32,

  std_in_handle: HANDLE,
  std_out_handle: HANDLE,
  pending_repeats: Option<(u16, TermEvent)>,

  std_in: ManuallyDrop<Stdin>,
  std_in_lock: ManuallyDrop<StdinLock<'static>>,

  std_out: ManuallyDrop<Stdout>,
  std_out_lock: ManuallyDrop<StdoutLock<'static>>,

  std_err: ManuallyDrop<Stderr>,
  std_err_lock: ManuallyDrop<StderrLock<'static>>,
}
impl Drop for WindowsTerminal {
  fn drop(&mut self) {
    unsafe {
      set_console_mode(self.std_in_handle, self.orig_console_input).unwrap();
      //
      set_console_mode(self.std_out_handle, self.orig_console_output).unwrap();
      //
      ManuallyDrop::drop(&mut self.std_in_lock);
      ManuallyDrop::drop(&mut self.std_in);
      ManuallyDrop::drop(&mut self.std_out_lock);
      ManuallyDrop::drop(&mut self.std_out);
      ManuallyDrop::drop(&mut self.std_err_lock);
      ManuallyDrop::drop(&mut self.std_err);
    }
  }
}
impl WindowsTerminal {
  pub fn new() -> Self {
    let std_in = stdin();
    let std_out = stdout();
    let std_err = stderr();
    let std_in_lock: StdinLock<'static> = unsafe { transmute(std_in.lock()) };
    let std_out_lock: StdoutLock<'static> =
      unsafe { transmute(std_out.lock()) };
    let std_err_lock: StderrLock<'static> =
      unsafe { transmute(std_err.lock()) };

    let std_in_handle = get_std_handle(StdHandle::Input).unwrap();
    let orig_console_input =
      unsafe { get_console_mode(std_in_handle).unwrap() };
    let new_console_input = ENABLE_WINDOW_INPUT;
    unsafe { set_console_mode(std_in_handle, new_console_input).unwrap() };

    let std_out_handle = get_std_handle(StdHandle::Output).unwrap();
    let orig_console_output =
      unsafe { get_console_mode(std_out_handle).unwrap() };
    let new_console_output =
    ENABLE_PROCESSED_OUTPUT | ENABLE_VIRTUAL_TERMINAL_PROCESSING | DISABLE_NEWLINE_AUTO_RETURN;
    unsafe { set_console_mode(std_out_handle, new_console_output).unwrap() };

    Self {
      orig_console_input,
      orig_console_output,

      std_in_handle,
      std_out_handle,
      pending_repeats: None,

      std_in: ManuallyDrop::new(std_in),
      std_in_lock: ManuallyDrop::new(std_in_lock),

      std_out: ManuallyDrop::new(std_out),
      std_out_lock: ManuallyDrop::new(std_out_lock),

      std_err: ManuallyDrop::new(std_err),
      std_err_lock: ManuallyDrop::new(std_err_lock),
    }
  }

  pub fn flush(&mut self) -> std::io::Result<()> {
    Ok(())
  }

  pub fn poll_events(&mut self) -> Result<Option<TermEvent>, i32> {
    if let Some((ref mut count, event)) = self.pending_repeats {
      if *count > 0 {
        *count -= 1;
        return Ok(Some(event));
      } else {
        self.pending_repeats = None;
      }
    }
    let mut record = INPUT_RECORD::default();
    loop {
      let count = unsafe {
        read_console_input(
          self.std_in_handle,
          core::slice::from_mut(&mut record),
        )
        .map_err(|e| e.0 as i32)?
      };
      if count > 0 {
        match record.EventType {
          KEY_EVENT => {
            let ker = unsafe { record.Event.KeyEvent() };
            if ker.bKeyDown == 0 || ker.wRepeatCount == 0 {
              continue;
            }
            let mods = {
              let mut mods = Mods::None;
              if ker.dwControlKeyState & (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED)
                > 0
              {
                mods.0 |= Mods::Alt.0;
              }
              if ker.dwControlKeyState
                & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED)
                > 0
              {
                mods.0 |= Mods::Ctrl.0;
              }
              if ker.dwControlKeyState & SHIFT_PRESSED > 0 {
                mods.0 |= Mods::Shift.0;
              }
              mods
            };
            let val = match i32::from(ker.wVirtualKeyCode) {
              fenestroj::winuser::VK_LSHIFT => continue,
              fenestroj::winuser::VK_RSHIFT => continue,
              fenestroj::winuser::VK_SHIFT => continue,
              fenestroj::winuser::VK_LCONTROL => continue,
              fenestroj::winuser::VK_RCONTROL => continue,
              fenestroj::winuser::VK_CONTROL => continue,
              fenestroj::winuser::VK_MENU /*alt*/ => continue,
              fenestroj::winuser::VK_ESCAPE => Key::Escape,
              fenestroj::winuser::VK_TAB => Key::Tab,
              fenestroj::winuser::VK_RETURN => Key::Enter,
              fenestroj::winuser::VK_LEFT => Key::ArrowLeft,
              fenestroj::winuser::VK_UP => Key::ArrowUp,
              fenestroj::winuser::VK_RIGHT => Key::ArrowRight,
              fenestroj::winuser::VK_DOWN => Key::ArrowDown,
              fenestroj::winuser::VK_HOME => Key::Home,
              fenestroj::winuser::VK_END => Key::End,
              fenestroj::winuser::VK_INSERT => Key::Insert,
              fenestroj::winuser::VK_DELETE => Key::Delete,
              fenestroj::winuser::VK_BACK => Key::Backspace,
              fenestroj::winuser::VK_PRIOR => Key::PageUp,
              fenestroj::winuser::VK_NEXT => Key::PageDown,
              fenestroj::winuser::VK_CLEAR => Key::Keypad5,
              fenestroj::winuser::VK_F1 => Key::F(1),
              fenestroj::winuser::VK_F2 => Key::F(2),
              fenestroj::winuser::VK_F3 => Key::F(3),
              fenestroj::winuser::VK_F4 => Key::F(4),
              fenestroj::winuser::VK_F5 => Key::F(5),
              fenestroj::winuser::VK_F6 => Key::F(6),
              fenestroj::winuser::VK_F7 => Key::F(7),
              fenestroj::winuser::VK_F8 => Key::F(8),
              fenestroj::winuser::VK_F9 => Key::F(9),
              fenestroj::winuser::VK_F10 => Key::F(10),
              fenestroj::winuser::VK_F11 => Key::F(11),
              fenestroj::winuser::VK_F12 => Key::F(12),
              fenestroj::winuser::VK_OEM_PLUS |
              fenestroj::winuser::VK_OEM_COMMA |
              fenestroj::winuser::VK_OEM_MINUS |
              fenestroj::winuser::VK_OEM_PERIOD |
              fenestroj::winuser::VK_OEM_1 |
              fenestroj::winuser::VK_OEM_2 |
              fenestroj::winuser::VK_OEM_3 |
              fenestroj::winuser::VK_OEM_4 |
              fenestroj::winuser::VK_OEM_5 |
              fenestroj::winuser::VK_OEM_6 |
              fenestroj::winuser::VK_OEM_7 |
              fenestroj::winuser::VK_OEM_8 |
              fenestroj::winuser::VK_OEM_102 |
              32..=126 => {
                Key::Lit(unsafe {*ker.uChar.AsciiChar()} as u8)
              }
              _ => {
                write!(self, "unknown ker:\r\n").unwrap();
                write!(self, "bKeyDown:{:?}\r\n", ker.bKeyDown).unwrap();
                write!(self, "wRepeatCount:{:?}\r\n", ker.wRepeatCount).unwrap();
                write!(self, "wVirtualKeyCode:{:?} (0x{:02X})\r\n", ker.wVirtualKeyCode,ker.wVirtualKeyCode).unwrap();
                write!(self, "dwControlKeyState:{:?}\r\n", ker.dwControlKeyState).unwrap();
                continue;
              },
            };
            let event = TermEvent::Key { val, mods };
            if ker.wRepeatCount > 1 {
              self.pending_repeats = Some((ker.wRepeatCount - 1, event));
            }
            return Ok(Some(event));
          }
          WINDOW_BUFFER_SIZE_EVENT => {
            let wbser = unsafe { record.Event.WindowBufferSizeEvent() };
            return Ok(Some(TermEvent::Resize {
              rows: wbser.dwSize.Y as u16,
              cols: wbser.dwSize.X as u16,
            }));
          }
          _ => {
            continue;
          }
        }
      } else {
        return Ok(None);
      }
    }
  }
}

impl Write for WindowsTerminal {
  #[inline]
  fn write_str(&mut self, s: &str) -> Result<(), core::fmt::Error> {
    // Note(Lokathor): It might seem insane to not buffer the output at all, but
    // testing has shown that it's sufficiently fast to just do this instead of
    // collecting into a Vec<u16> or anything like that.
    let mut buf = [0_u16, 0_u16];
    for ch in s.chars() {
      write_console(self.std_out_handle, ch.encode_utf16(&mut buf))
        .map_err(|_| core::fmt::Error)?;
    }
    Ok(())
  }
}
